---------------------------------
-- REX 3d-mesh file format (.RMesh)
-- Author: Rex Hill (rex3dworks@hotmail.com)
--

g_bf2doInclude "general/misc.ms"

struct BF2GetSkinWeights_class
(
	ids=#(),
	weights=#()
)




-----------------------------------
--- Skinning
--
-- skeletonBones: an array of nodes since this skin modifier only returns names
--
fn BF2GetSkinWeights obj bnObjs dontWarn:true=
(
	local skinMod = obj.modifiers[#skin]
	if skinMod == undefined then return #(#(),#())
		
	format "BF2GetSkinWeights: %\n" (BF2getObjectPath obj)
	
	local bnNames = #()
	for b in bnObjs do
		append bnNames (lowercase b.name)
		
	-- necessary for skinOps to work
	max modify mode
	select obj

	local boneUsed = #()
	local vWeights = #()
	
	skinMod.bone_Limit = 2 -- only 2 bones affect one vert currently in game
	
	local numVerts = skinOps.GetNumberVertices skinMod
	
	for i=1 to numVerts do
	(
		local itVertWeights = BF2GetSkinWeights_class #(-1,-1,-1,-1) #(0,0,0,0)
		
		local WeightCnt = skinOps.GetVertexWeightCount skinMod i
		
		for j=1 to weightCnt do
		(
			if j > 4 then exit
			
			local IDa = SkinOps.GetVertexWeightBoneID skinMod i j
			local bnName = lowercase (SkinOps.GetBoneName skinMod IDa 0)
			local bnId = (findItem bnNames bnName)
			boneUsed[bnId] = true
			itVertWeights.ids[j] = bnId - 1
			itVertWeights.weights[j] = skinOps.GetVertexWeight skinMod i j
		)
		
		vWeights[i] = itVertWeights
	)
	
	-- run through again, this time we know the final bone ids (because we want to remove those bones not used)
	
	local condensedBoneList = #()

	for i=1 to bnObjs.count do
	(
		if boneUsed[i] == true then
		(	
			append condensedBoneList bnObjs[i]
		)
	)
	
	local bnNames2 = #()
	for b in condensedBoneList do
	(	
		append bnNames2 (lowercase b.name)
	)

	for i=1 to numVerts do
	(
		local WeightCnt = skinOps.GetVertexWeightCount skinMod i
		
		for j=1 to weightCnt do
		(
			if j > 4 then exit
			
			local IDa = SkinOps.GetVertexWeightBoneID skinMod i j
			local bnName = lowercase (SkinOps.GetBoneName skinMod IDa 0)
			local bnId = (findItem bnNames2 bnName)
			vWeights[i].ids[j] = bnId - 1
		)
-- added error handling to show unweighted vertex mschoeldgen Oct 2006
        if vWeights[i].ids[1] == -1 then 		
            if vWeights[i].ids[2] == -1 then 		
                if vWeights[i].ids[3] == -1 then 		
                    if vWeights[i].ids[4] == -1 then 		
							format "ERROR No weight in vertex %\n" (i-1)
	)

	
	--format "oldBnList: %\n" bnObjs
	--format "condensedBoneList: %\n" condensedBoneList 
--for v in vWeights do format " %\n" v
		
	if not dontWarn then
	(
		local hadError = false
		for v in vWeights do
		(
			if v.ids[1] == -1 then
				if v.ids[2] == -1 then
					if v.ids[3] == -1 then
						if v.ids[4] == -1 then
						(
							format "ERROR No weight! %\n" v
							hadError  = true
						)
		)
		
		if hadError == true then
		(
			MessageBox "ERROR! Atleast one vertex was not weighted to any bone in the proper skeleton!\n Export will likely crash!"
		)
	)
	if condensedBoneList.count > 26 then
	(
		MessageBox "ERROR! More than 26 skeleton bones are used on this mesh!"
	)
	
	return #(vWeights, condensedBoneList )
)








-----------------
-- Sets all smoothgroups to 1
-- clears all UVW mapping information
--
mapped fn bf2prepColMesh obj =
(
	local opClass = undefined
	local objClass = classof obj
	case objClass  of
	(
		Editable_Poly: opClass = polyOp
		Editable_Mesh: opClass = meshop
		TriMesh: opClass = meshop
		default: opClass = undefined
	)
	

	if opClass == undefined then return false
	
	local faceCnt = opClass.getNumFaces obj
	
	if objClass == Editable_Poly then
	(
		for i=1 to faceCnt do
		(
			polyOp.setFaceSmoothGroup obj i 1
		)
		
		for i=0 to (polyOp.getNumMaps obj) do
		(
			polyOp.setMapSupport obj i false
		)
		
		-- the resize function doesn't work
		-- would have to add UVW mapping Clear modifiers
	)
	else
	(
		for i=1 to faceCnt do
		(
			setFaceSmoothGroup obj i 1
		)
		
		opClass.setNumMaps obj 2 keep:false
		opClass.setMapSupport obj 1 false	
		opClass.setMapSupport obj 0 false
			
	)


	return true
)











struct RMesh 
(
--------------------
	fn GetVer = return 17,
	fn GetFType = return "RMesh(*.RMesh)|*.RMesh|All|*.*|",
	
	fn writePoint3 f pnt3 = (writeFloat f pnt3.x; writeFloat f pnt3.y; writeFloat f pnt3.z;),
	fn writePoint3i f pnt3 = (writeLong f pnt3.x; writeLong f pnt3.y; writeLong f pnt3.z;),
	
	fn readPoint3 f = ( return [readFloat f, readFloat f, readFloat f] ),
	fn readPoint3i f = ( return [readLong f #unsigned, readLong f #unsigned, readLong f #unsigned] ),

--------------------
-- IMPORT ----------
--------------------
	fn import fName:undefined s:10.0=
	(
		if fName == undefined then fName = getOpenFileName types:( RMesh.GetFType() )
		if fName == undefined then return();
		
		format "Importing file: %\n" fName
		local f = fopen fName "rb"
		if f != undefined then
		(
		undo off
		(
	
			local header = ReadLong f
			if header != (RMesh.GetVer()) then
			(
				messagebox ("ERROR! Unknown header type:\n" +(header as string))
				
				fclose f
				return false
			)
		
			local tmpPt3
			
			-- pos
			local theVerts = #()
			local numVerts = readlong f #unsigned
			for i=1 to numVerts do
			(
				tmpPt3 = (RMesh.readPoint3 f) * s
				theVerts[i] = [tmpPt3.x, tmpPt3.z, tmpPt3.y]
			)
			
			-- normals (ignored by 3dsmax so don't import unless debugging)
			local theVertNormals = #()
			local numVertNormals = readlong f #unsigned
			for i=1 to numVertNormals do
			(
				tmpPt3 = RMesh.readPoint3 f
				-- theVertNormals[i] = tmpNorm
			)
			
			
			-- colors
			local numVertColors = readlong f #unsigned
			for i=1 to numVertColors do
			(
				tmpPt3 = RMesh.readPoint3 f -- r, g, b
				tmpPt3 = readFloat f -- alpha
			)
			
			-- weights
			local numVertWeights = readlong f #unsigned
			for i=1 to numVertWeights do
			(
				-- vertex pos for identification
				tmpPt3 = RMesh.readPoint3 f 
				
				-- Bone IDs
				for i=1 to 4 do
					tmpPt3 = readlong f #unsigned
					
				-- weights
				for i=1 to 4 do
					tmpPt3 = readFloat f #unsigned			

			)
			
			local theFaces = #()
			local numFaces = readLong f #unsigned			
			for i=1 to numFaces do
			(
				tmpPt3 = RMesh.readPoint3i f
				theFaces[i] = [ tmpPt3.z, tmpPt3.y, tmpPt3.x ] + 1
			)
	
			format "Header: %\tNumVerts: %\tNumFaces: %\n" header numVerts numFaces
			
			local theFacesMatID = #()
			local numMatIds = readLong f #unsigned
			for i=1 to numMatIds do	
				theFacesMatID[i] = readLong f #unsigned

			local tmpMeshNode = mesh vertices:theVerts faces:theFaces materialIDs:theFacesMatID
			
			local numSmoothGroups = readLong f #unsigned
			local tmpSmoothGroupInt = 0
			for i=1 to numSmoothGroups do
			(			
				tmpSmoothGroupInt = readLong f #unsigned
				setFaceSmoothGroup tmpMeshNode i tmpSmoothGroupInt
			)
			
			
			-- Texture Info Here
			
			local numMapChannels = readlong f #unsigned
			meshop.setNumMaps tmpMeshNode (numMapChannels+1)
			
			format "numMapChannels: %\n" numMapChannels
			for n=1 to numMapChannels do
			(	
				local theMapVerts = #()
				local numMapVerts = readlong f #unsigned
				format " numMapVerts: %\n" numMapVerts
				for i=1 to numMapVerts do
					theMapVerts[i] = RMesh.readPoint3 f
	
	
				local theMapFaces = #()
				local numMapFaces = readlong f #unsigned
				format "   numMapFaces: %\n" numMapFaces
				for i=1 to numMapFaces do
				(
					theMapFaces[i] = RMesh.readPoint3i f
					theMapFaces[i] += 1
				)
		
				if numMapVerts > 0 then
				(
					meshop.setNumMapFaces tmpMeshNode n numMapFaces
					meshop.setNumMapVerts tmpMeshNode n numMapVerts

					for i=1 to numMapVerts do
						meshop.setMapVert tmpMeshNode n i theMapVerts[i]
						
					for i=1 to numMapFaces do
						meshop.setMapFace tmpMeshNode n i theMapFaces[i]
				)
				
			)

			fclose f
			return tmpMeshNode;

		)
		)
		--else messagebox ("ERROR! Could not open:\n" +(fName as string))
		
		return undefined
	),
	
	
	
	
--------------------
-- EXPORT ----------
--------------------
	-- materialStruct: u32 NumTextures, u16 textureNameLn, char[] textureName (null terminated)
	-- getTextureName
	fn export objA fHandle:undefined fname:undefined forceMatID:-1 s:0.1 coordSpace:#world matInfo:undefined isNonVis:false bnObjs:#() dontWarnSkinWeights:false minUvSetCnt:0 lmChan:9=
	(
	
		--format "mesh: % - minUvSetCnt: %\n" objA.name minUvSetCnt
		local maxVertWeights = 4 -- max num bones affecting one vertex
		local vertWeights = #()
		
		--format "RMESH EXPORT()\n"
		--format " bnObjs: %\n" bnObjs
		
		--format "%\n" objA 

		-- coordSpace can be:
			-- a node in the scene
			-- a transformation value (matrix3)
			-- all others mean coordSpace = worldspace coordinates

		local obj = undefined
		
		-- check that the geometry is valid
		if not (isValidNode objA) then return false;
		if SuperClassof objA != GeometryClass then return false;
		try
		(
			--format "  coordSpace: % %\n" coordSpace (classof coordSpace)
			-- modify the mesh's transform temporarily to retrieve the mesh data
			
			local oldTransform = objA.transform
			if isValidNode coordSpace then
			(
				objA.transform = objA.transform * (inverse coordSpace.transform)
			)
			else
			(
				if classof coordSpace == Matrix3 then
					objA.transform = coordSpace
			)
	
			-- this stores the mesh in world space as it appears
			obj = snapShotAsMesh objA
			
			
			if bnObjs.count > 0 then
			(
				
				local tmp = BF2GetSkinWeights objA bnObjs dontWarn:dontWarnSkinWeights
				vertWeights = tmp[1]
				bnObjs = tmp[2]
				
				if tmp[2].count > 0 then
				(
					format " bnObjs (used): %\n" bnObjs
				)
			)
			
			-- restore object's original position
			objA.transform = oldTransform

			if isNonVis then
				bf2prepColMesh obj			
		)
		catch 
		(	return false
		)
	
		
		-- Open File for writing
		local f = fHandle
		if f == undefined then
		(
			if fName == undefined then fName = getSaveFileName types:(RMesh.GetFType())
			if fName == undefined then return false;
				
			format "Saving: %\n" fName
			f = fopen fName "wb"		
		)		

		if f != undefined then
		(
			undo off
			(
				writeLong f (RMesh.GetVer()) -- header

				local pnt3
				
			--	meshop.deleteIsoMapVertsAll obj
				
				-- Verts Pos
				writeLong f (obj.numVerts) #unsigned
				for i=1 to obj.numVerts do
				(
					pnt3 = (getVert obj i) * s
					writeFloat f pnt3.x
					writeFloat f pnt3.z
					writeFloat f pnt3.y
					
				--	format "  %\n" pnt3
				)
	
				
				-- Vert Normals
				writeLong f 0
				-- Re-calculate by smoothgroup later	
				-- keep support in the file incase exporter is needed from different 3d application		
	
				-- vert colors
				if (meshop.getMapSupport obj 0) then
				(	
					format "  HasVertexColors: %\n" (BF2getObjectPath objA)
					-- meshop.deleteIsoMapVerts obj 0
					
					writeLong f (obj.numVerts) #unsigned
					
					if meshop.getNumMapVerts obj 0 < obj.numVerts then
					(
						throw ("RMESH.export() ERROR! There are less vertex colors than there are verticies!\n OBJECT: " + (BF2getObjectPath objA))
					)
					else
					(
						for i=1 to obj.numVerts do
						(
							pnt3 = meshop.getMapVert obj 0 i
	
							if pnt3 == [1,1,1] or pnt3 == [0,0,0] then
								pnt3 = [-1,-1,-1]
	
							--format "pnt3: %\n" pnt3 
							
							writeFloat f pnt3.x
							writeFloat f pnt3.y
							writeFloat f pnt3.z
							writeFloat f 1.0
						)
					)
				)
				else
					writeLong f 0 #unsigned


				local numVertWeights = VertWeights.count
				
				if numVertWeights > 0 and numVertWeights != obj.numVerts then 
				(	
					format "ERROR! VertWeights(%) != numVerts(%)\n" vertWeights.count obj.numVerts
					fclose f
					return false
				)
				else
				(
					if numVertWeights > 0 then
						format "   numVertWeights: %\n" numVertWeights 
						
					writeLong f numVertWeights 
					for i=1 to numVertWeights do
					(			
						-- position to identify this vert after vert duplication/processing of the mesh		
						pnt3 = (getVert obj i) * s
						writeFloat f pnt3.x
						writeFloat f pnt3.z
						writeFloat f pnt3.y
						
						local vWeights = vertWeights[i]
						
						for j=1 to maxVertWeights do
							writeLong f vWeights.ids[j]
					
						for j=1 to maxVertWeights do
							writeFloat f vWeights.weights[j]
	
					)
				)
				
				
				-- Faces
				format "obj.numFaces: %\n" obj.numFaces

				writeLong f (obj.numFaces) #unsigned
				for i=1 to obj.numFaces do 
					RMesh.writePoint3i f ((getFace obj i) - 1)


				-- MatIDs
				(			
					writeLong f (obj.numFaces) #unsigned
									
					if matInfo != undefined then
					(	
						-- check if object has more material ID's than materials
						local materialMatIDs = BF2MatInfo.getMatIds matInfo
						if materialMatIDs.count == 1 then
						(
							forceMatID = 1;
						)
					)
					
					-- check that object has a material
								
					--format "  forceMatID: % - materialMatIDs: %\n" forceMatID materialMatIDs 

					
					if (forceMatID == -1) then
					(
						for i=1 to obj.numFaces do
							writeLong f (getFaceMatID obj i)
					)
					else
					(
						for i=1 to obj.numFaces do
							writeLong f forceMatID	
					)
					
					
				)

				-- SmoothingGroups
				writeLong f (obj.numFaces) #unsigned
				for i=1 to obj.numFaces do
					writeLong f (getFaceSmoothGroup obj i)


				-- Texture information
				local mapChannelsMap = #()
				for i=1 to (meshop.getNumMaps obj) do
				(
					if (meshop.GetMapSupport obj i) then
						append mapChannelsMap i
				)
				
				-- This makes a staticmesh always have 4 uvsets + lightmap (if there is a lightmap channel)
				--  the first UV set is copied to all missing uv sets
				-- note: this does not make dirt always channel 3 and crack always channel 4, that is unrelated to this
				--  this is only to ensure lightmap channels remain the same in the exported mesh (channel 5)
				if minUvSetCnt > 0 then
				(
					format " minUvSetCnt!: %\n" minUvSetCnt 
					
					local hasLmChan = false
					if meshop.getMapSupport obj lmChan then
					(
						hasLmChan = true
						deleteItem mapChannelsMap mapChannelsMap.count
					)
	
					if mapChannelsMap.count < minUvSetCnt then
					(
						local originalCount = mapChannelsMap.count + 1
						local refChannel = mapChannelsMap[1]
						
						if refChannel == undefined then
						(
							-- build uv coordinates for this object!
							refChannel = 1
							meshop.setMapSupport obj 1 true
						)
						
						for i=originalCount to minUvSetCnt do
						(
							append mapChannelsMap refChannel
						)
					)
				
					if hasLmChan then
					(
						append mapChannelsMap lmChan 
					)

				)
				
				writeLong f mapChannelsMap.count #unsigned
				for n=1 to mapChannelsMap.count do
				(
					local c = mapChannelsMap[n]

					local tmpNumMapVerts = 0
					if classof c == integer then
						tmpNumMapVerts = meshop.getNumMapVerts obj c
						
					writeLong f tmpNumMapVerts
					for i=1 to tmpNumMapVerts do
					(
						pnt3 = meshop.getMapVert obj c i
						writeFloat f pnt3.x
						writeFloat f pnt3.y
						writeFloat f pnt3.z
					)
					
					local tmpNumMapFaces = 0
					if classof c == integer then
						tmpNumMapFaces = meshop.getNumMapFaces obj c	
					
					writeLong f tmpNumMapFaces 
					for i=1 to tmpNumMapFaces do
					(
						RMesh.writePoint3i f ((meshop.getMapFace obj c i) - 1)
					)
				)
				
				writeLong f bnObjs.count
				for b in bnObjs do
				(
					local pathStr = BF2getObjectPath b
					writeLong f (pathStr.count + 1)
					writeString f pathStr
				)

				if fHandle == undefined then
					fclose f
			)
			return true
		)
		else messageBox ("ERROR! Could not create:\n" +(fName as string))
		
		return false
	)
	
-- end Struct
)






--------------------
-- TEST
/*
fn test1 =
(
	ClearListener()
	local f = fopen "c:\\test1.rMesh" "wb"
	
	RMesh.Export selection[1] coordSpace:(matrix3 [1,0,0] [0,1,0] [0,0,1] [0,0,0]) s:0.1 fHandle:f  \
		matInfo:#(matContainer materialID:1 materialName:"07_-_Default" fxFilename:undefined technique:undefined textures:#()) \
		isNonVis:False 
	
fclose f
	
--	objA: $Editable_Mesh:mesh2 @ [1.470968,5.147552,12.721709] 
--fHandle:<BinStream:C:\g\mods\bf2\Objects\TESTANM\teapot\teapot.dat> 
--fname:undefined forceMatID:-1 s:0.1 coordSpace:(matrix3 [1,0,0] [0,1,0] [0,0,1] [0,0,0]) 
-- matInfo:#(matContainer materialID:1 materialName:"07_-_Default" fxFilename:undefined technique:undefined textures:#()) 
--isNonVis:false

)


(
	ClearListener()
	try delete geometry[2]
	catch()
	
	local timeStart = timeStamp()
	 
	-- RMesh.Export geometry[1] fname:"c:\\4v_5t_2f.rMesh"
	local tmpMesh = RMesh.Import fname:"c:\\output.rMesh"

	tmpMesh.material = geometry[1].material
	
	format "EndTime: %\n" (timeStamp() - timeStart)
 
)
*/

